/***************************************************************************
                          sorcery.c  -  for sorcery and GUI--
                             -------------------
    begin                : Tue Sep 11 2001
    copyright            : (C) 2001 by Josiah Zayner
    email                : phric@legions.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "wand.h"



void ip_sorcery(struct ip *ip_sorc,
                struct in_addr source,
                struct in_addr dest,
                int proto,
                unsigned int hl,
                unsigned int ver,
                unsigned char tos,
                unsigned short len,
                unsigned short id,
                unsigned char ttl,
                int off
               )
{
  ip_sorc->ip_hl = hl;
  ip_sorc->ip_v = ver;
  ip_sorc->ip_tos = tos;
  ip_sorc->ip_len = len;
  ip_sorc->ip_id = id;
  ip_sorc->ip_ttl = ttl;
  ip_sorc->ip_p = proto;
  ip_sorc->ip_off = off;
  ip_sorc->ip_src = source;
  ip_sorc->ip_dst = dest;

}

int mix_ipoption(unsigned int opt,
                 unsigned int optlen,
                 unsigned int totlen,
                 struct ip *ip_p,
                 char *pack
                )
{
  static int oper;
  /* optlen = x8bits  we need to ocnvert to x32bits */
  if(optlen % 4){ oper = (optlen / 4) + (optlen % 4); }
  else { oper = (optlen / 4); }
  ip_p->ip_len = totlen;
  ip_p->ip_hl = 5 + oper; /* hl is in x32 bits */
  memcpy(pack, ip_p, S_IP);
  memcpy(pack + S_IP, &opt, (oper * 4));

  return (oper * 4);
}

int tcp_sorcery(struct tcphdr *tcp_magic,
                struct ip *ip_magic,
                char *datap,
                int dps,
                int ipopt,
                int optlen,
                int num
               )
{
  char *pack = clalloc(65535);
  char *envelope = clalloc(S_NMR + S_TCP + dps);
  struct nmrhdr *nmr_magic = (struct nmrhdr *)envelope;
  static int totlen;
  static int a = 0;


  /* create psuedo header */
  nmr_magic->saddr = ip_magic->ip_src;
  nmr_magic->daddr = ip_magic->ip_dst;
  nmr_magic->none = 0;
  nmr_magic->protocol = 6;
  nmr_magic->length = htons(S_TCP + dps);
  /* putting tcp header and data in packet for checksum */
  memcpy(envelope + S_NMR, tcp_magic, S_TCP);
  memcpy(envelope + S_NMR + S_TCP, datap, dps);

  /* calculate checksum for the psuedo no miss routed packet header and tcp header */
  tcp_magic->th_sum = in_cksum((u_short *)envelope, S_TCP + S_NMR + dps);
  free(envelope);

  totlen = S_TCPIP + dps;
  optlen = mix_ipoption(ipopt, optlen, totlen, ip_magic, pack);
  /* putting our tcp header in the final packet */
  memcpy(pack + S_IP + optlen, tcp_magic, S_TCP);
  memcpy(pack + S_TCPIP + optlen, datap, dps);

  for(; num > a; a++)
  {
    /* send packet */
    if(cast_spell(pack,
                  (S_TCPIP + dps + optlen),
                  tcp_magic->th_dport,
                  ip_magic->ip_dst)
                  == -1)
    {
       perror("sendto");
       return -1;
    }
  }
  free(pack);
  return 0;
}


int udp_sorcery(struct udphdr *udpp,
                struct ip *ip_magic,
                char *datap,
                int dps,
                int ipopt,
                int optlen,
                int ipopte,
                int num
               )
{
  char *pack = clalloc(65535);
  char *envelope = clalloc(S_NMR + S_UDP + dps);
  struct nmrhdr *nmr_magic = (struct nmrhdr *)envelope;
  static int totlen, a = 0;

  nmr_magic->saddr = ip_magic->ip_src;
  nmr_magic->daddr = ip_magic->ip_dst;
  nmr_magic->none = 0;
  nmr_magic->protocol = 17;
  nmr_magic->length = htons(S_UDP + dps);

  memcpy(envelope + S_NMR, udpp, S_UDP);
  memcpy(envelope + S_NMR + S_UDP, datap, dps);

  /* calculate checksum for the psuedo no miss routed packet header + udp header + data */
  udpp->uh_sum = in_cksum((u_short *)envelope, S_UDP + S_NMR + dps);
  free(envelope);

  totlen = S_IP + optlen + S_UDP + dps;
  optlen = mix_ipoption(ipopt, optlen, totlen, ip_magic, pack);

  /* putting our udp header in the packet */
  memcpy(pack + S_IP + optlen , udpp, S_UDP);
  memcpy(pack + S_IP + optlen + S_UDP, datap, dps);

  for(; num > a; a++)
  {
    if(cast_spell(pack,
                  (S_IP + optlen + S_UDP + dps),
                  udpp->uh_dport,
                  ip_magic->ip_dst) == -1)
    {
       perror("sendto");
       return -1;
    }
  }

  return 0;

}



int igmp_sorcery(struct igmp *igmp_magic,
                 struct ip *ip_magic,
                 char *datap,
                 int dps,
                 int ipopts,
                 int optlen,
                 int num
                )
{


  char *pack = clalloc(65535);
  static int totlen, a = 0;

  /* for IGMP checksum */
  memcpy(pack, igmp_magic, sizeof(struct igmp));
  memcpy(pack + sizeof(struct igmp), datap, dps);

  /* calculate checksum of IGMP + data */
  igmp_magic->cksum = in_cksum((u_short *)pack, sizeof(struct igmp) + dps);

  totlen = S_IP + optlen + S_UDP + dps;
  mix_ipoption(ipopts, optlen, totlen, ip_magic, pack);

  /* calculate IP header checksum */
  memcpy(pack + S_IP + optlen, igmp_magic, sizeof(struct igmp));
  memcpy(pack + S_IP + sizeof(struct igmp) + optlen, datap, dps);


  for(; num > a; a++)
  {
    if(cast_spell(pack, S_IP + sizeof(struct igmp) + dps + optlen, 0, ip_magic->ip_dst) == -1)
    {
       perror("sendto");
       return -1;
    }
  }
  free(pack);
  return 0;
}

int cast_spell(char *envelope, int env_size, unsigned char deport, struct in_addr src)
{
  int s, yes = 1;
  struct sockaddr_in endup;

 /* open a raw socket */
  if((s = socket(PF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
  {
    perror("socket");
    exit(s);
  }

  /* set socket options at IP level, so we can create our own IP header */
  if((setsockopt(s, SOL_IP, IP_HDRINCL, &yes, sizeof(yes))) == -1)
  {
    perror("setsockopt");
    exit(-1);
  }

  /* fill in sockaddr struct bLAH! */
  memset(&endup, '\0', sizeof(struct sockaddr_in));
  endup.sin_family = PF_INET;
  endup.sin_port = deport;
  endup.sin_addr = src;

  /* send packet woohooo! */
  sendto(s, envelope, env_size, 0, (struct sockaddr *)&endup, sizeof(struct sockaddr));

  return s;
}


void prterr(char *message)
{ fprintf(stderr, message); }

